/*$Id: rand.c,v 1.6 2002/12/09 10:42:08 aa5779 Exp $*/
/*
 	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <modules.h>
#include <util.h>
#include <time.h>
#include <dstring.h>
#include "version.h"
//// \author[Artem V. Andreev]
//// \copyright[2002 Artem V. Andreev -- See \ref[sl(1)]]
/// \section[NAME]
// \em[rand] -- a module for Tensile (see \ref[sl(1)])

/// \section[OVERVIEW]
//// This module provides access to random number generators

static module *this_mod;
extern procstack *get_pstack(void);
extern void define_var(interpreter *i, int status, char *name, dstring *val);

void rand_LTX_init(module *this)
{
	this_mod = this;
}

//// \section[FUNCTIONS]
//// rand()
//// \returns the next pseudorandom value of a software generator (see \ref[rand(3)])
static void bi_rand(chunk *current, interpreter *i, value *args)
{
	ds_append(get_pstack()->state->buffer, ds_fromint(rand(), 10, 0));
}

//// srand([seed])
//// Sets the seed of a software random generator to \em[seed],
////  or to the current time value if \var[seed] is not present
static void bi_srand(chunk *current, interpreter *i, value *args)
{
	srand(args ? convert_to_int(DS_BODY(args->val)) : time(NULL));
}

#if defined(HAVE__DEV_RANDOM) || defined(HAVE__DEV_URANDOM)
static int num_to_hex(int num)
{
	return num < 10 ? num + '0' : num + 'A' - 10;
}

//// highrand(n strict)
//// \returns \var[n] bytes of highly-randomized (probably hardware-generated) value
//// If \var[strict] is empty, a less reliable generator may be used
//// \note On Linux (as well as some other *nix systems) this functions accesses
//// \code[/dev/random] or \code[/dev/urandom] resp.
//// \note This function will not be present, if no adequate generator is available
static void bi_highrand(chunk *current, interpreter *i, value *args)
{
#ifdef HAVE__DEV_RANDOM
#ifdef HAVE__DEV_URANDOM
	FILE *f = fopen(ds_isempty(next_value(args)->val) ? "/dev/random" : "/dev/urandom", "rb");
#else
	FILE *f = fopen("/dev/random", "rb");
#endif
#else
	FILE *f = fopen("/dev/urandom", "rb");
#endif
	int len = convert_to_int(DS_BODY(args->val));
	int ch;

	if(len <= 0)
		recover_error("invalid length (%d) for 'highrand'", len);
	if(!f)
		recover_error("can't open random generator devices");
	while(len--)
	{
		while((ch = getc(f)) == EOF) // in case /dev/random should return EOF when no data available
			;
		ds_appendch(get_pstack()->state->buffer, num_to_hex(ch >> 4));
		ds_appendch(get_pstack()->state->buffer, num_to_hex(ch & 0x0F));
	}
	fclose(f);
}
#endif

static master_instruction_src rand_src_add[] = {
										 { "rand", 1, bi_rand, 0, 0, 0},
										 { "srand", 0, bi_srand, 0, 1, 1},
#if defined(HAVE__DEV_RANDOM) || defined(HAVE__DEV_URANDOM)
										 { "highrand", 1, bi_highrand, 2, 2, 2}, 
#endif
										 {NULL}
										};

void rand_LTX_attach(mod_attach *ma)
{
	variable *v;
	append_builtins(ma->i, rand_src_add);
	define_var(ma->i, var_global, "RAND_MAX", ds_fromint(RAND_MAX, 10, 0));
	v = safe_get_var(ma->i, "RAND_MAX", 0);
	v->readonly = 1;
}

const char rand_LTX_version[] = VERSION;
const char rand_LTX_date[] = __DATE__;
const char rand_LTX_srcdate[] = "$Date: 2002/12/09 10:42:08 $";

/// \trailer
